;
; Play 8bit DMA mode for SoundBlaster v1.00
;   Andr Baresel (with some help from Craig Jackson)
;
; STATUS: DOES WORK ON SB16/SBPRO2/SB2.0, TESTPHASE ON SB1.0
;
; Requirements: 80286, SoundBlaster (see BASEADDR,DMA channel,IRQ number)
; Resolutions : 8-bit / 4..23khz (no highspeed - but on SB16 4..44kHz)
; Parameters  : none
; Notes:
;   To creat a 8 bit mono unsigned file do :   "VOC2RAW TEST1.VOC /I"
;
;  DSP command 14h  ... play 8bit mono no autoinit
;  DSP command 40h  ... set sample rate
;  DSP command D1h  ... Enable Speaker
;  DSP command D3h  ... Disable Speaker
;

	.386
	.MODEL FLAT

; CONSTANTS 

; SoundBlaster SETUP
BASEADDR           EQU 0220h       ;SoundBlaster base address
SBIRQ              EQU 7           ;SoundBlaster IRQ
DMAchannel         EQU 1           ;SoundBlaster DMA channel

; PIC MASKS FOR MASK/DEMASK IRQ
PICORMASK          EQU 1 shl SBIRQ
PICANDMASK         EQU 0FFh xor PICORMASK

; DMA CONTROLLER REGISTERS :
WRITEMASK          EQU 00ah         ;WRITE MASK REGISTER
WRITEMODE          EQU 00bh         ;WRITE MODE REGISTER
CLEARFLIPFLOP      EQU 00ch
PAGE_CHN           EQU 083h         ;PAGE REGISTER FOR DMAchannel 1
BASE_CHN           EQU 002h         ;BASEADDRESS REGISTER DMA 1
COUNT_CHN          EQU 003h         ;COUNT REGISTER DMAchannel 1

; SAMPLERATE : (if you change it pay attention to maximum samplerate)
TIMECONST          EQU 165          ; = 10989 Hz (256-1000000/11000)

; DMA WRITE MODE
WANTEDMODE         EQU 01001000b    ; singlemode, nonautoinit, readmode

;
; MACRO DEFINITIONs
;

WAITWRITE               MACRO
LOCAL                   loopWait,endloop
;          Arguments : DX = Status port (BASEADDR+0Ch)
;          Returns   : n/a
;          Destroys  : AL

                        push    ecx
                        mov     ecx,1000h       ; need that for slow SBs !
loopWait:               dec     ecx
                        jz      endloop
                        in      al,dx           ; AL = WRITE COMMAND STATUS
                        or      al,al
                        js      loopWait        ; Jump if bit7=1 - writing not allowed
endloop:                pop     ecx
ENDM

WAITREAD                MACRO
LOCAL                   loopWait,endloop
;          Arguments : DX = Status port   (normaly BASEADDR+0Eh)
;          Returns   : n/a
;          Destroys  : AL

                        push    ecx
                        mov     ecx,1000h       ; need that for slow SBs !
loopWait:               dec     ecx
                        jz      endloop
                        in      al,dx           ; AL = DATA AVAILABLE STATUS
                        or      al,al
                        jns     loopWait        ; Jump if bit7=0 - no data available
endloop:                pop     ecx
ENDM

RESET_DSP               MACRO
local                   SBthere
;          Arguments : n/a
;          Returns   : n/a
;          Destroys  : DX,AL

                        mov      dx,BASEADDR+06h
                        mov      al,1
                        out      dx,al          ; start DSP reset

                        in       al,dx
                        in       al,dx
                        in       al,dx
                        in       al,dx          ; wait 3 sec

                        xor      al,al
                        out      dx,al          ; end DSP Reset

                        add      dx,08h         ; dx = DSP DATA AVAILABLE
                        WAITREAD
                        sub      dx,4           ; dx = DSP Read Data
                        in       al,dx
                        cmp      al,0aah        ; if there is a SB then it returns 0AAh
                        je       SBthere
                        jmp      RESET_ERROR    ; No SB - exit program
SBthere:
ENDM
; End of Macrodefinitions 

	.DATA
;
; TWO COPIES FOR PAGE OVERRIDE REASONS :


	dataseg dd 0
	pBuffer dd 0

    ready               db 0

    information         db 'DMASTEP1.EXE - play a 8bit mono sample one time',13,10
                        db '         (use normal speed and nonautoinit DMA)',13,10
                        db 'stop playing with any key ...','$'

    sberror             db 'No SoundBlaster at this BASEADDR ! PROGRAM HALTED.',13,10,'$'
    dosmemerror         db 'Not enough DOS memory for sound data.',13,10,'$'

    oldInterruptOFS     dd 0
    oldInterruptSEG     dw 0

;

	.const
SAMPLEBUFFER LABEL BYTE
     INCLUDE TEST1.INC        ; COPY OF SAMPLE SOUND
SAMPLEBUFFEREND LABEL BYTE

SAMPLEBUFFERLENGTH equ offset SAMPLEBUFFEREND - offset SAMPLEBUFFER

.CODE

main proc

           mov [dataseg], ds

           RESET_DSP

           ; WRITE INFORMATION TO SCREEN :
           mov     edx,offset information
           mov     ah,9
           int     21h                         ; write program information to screen

		mov ebx, SAMPLEBUFFERLENGTH*2
		add ebx, 16-1
		shr ebx, 4
		mov ax, 100h
		int 31h
		jc dosmemerr
		movzx eax, ax
		shl eax, 4
		mov pBuffer, eax


           ; ENABLE SB SPEAKERS (for all SBs <SB16)
           mov     dx,BASEADDR+00Ch            ;DX = DSP Write Data or Command
           WAITWRITE
           mov     al,0D1h                     ; AL = Enable speaker
           out     dx,al                       ; Output: DSP Write Data or Command

           ; DISABLE IRQ
           in      al,021h
           or      al,PICORMASK                ; SET MASK REGISTER BIT TO DISABLE INTERRUPT
           out     021h,al

           ; CHANGE POINTER IN INTERRUPT TABLE
		   mov bl, SBIRQ
		   mov ax, 204h
		   int 31h
           mov [OLDInterruptOFS],edx
           mov [OLDInterruptSEG],cx
		   mov cx, cs
		   mov edx, offset OWN_IRQ
		   mov ax, 205h
		   int 31h

           ; CHANGE PIC MASK :
           in      al,021h
           and     al,PICANDMASK   ; CLEAR MASK REGISTER BIT TO ENABLE INTERRUPT
           out     021h,al

;
; calculate page and offset for DMAcontroller :
;
; segment*16+offset - 20bit memory location -> upper 4 bits  = page
;                                              lower 16 bits = offset
;
           mov     esi,pBuffer
           mov     ecx,SAMPLEBUFFERLENGTH

;
; check for DMApage override :
; ... problem: DMA controller separates memory into 64KB pages, you can only
; transfer data is placed in one page - no page overrides are allowed
;
; To solve that :
; creat a DMA buffer with double size you want - if the first part is placed
; on a page border the second part is for sure not
;
           lea eax, [esi+ecx-1]
		   mov ebx, esi
		   shr eax, 16
		   shr ebx, 16
		   cmp eax, ebx
		   jz nooverride
		   mov esi, eax
		   shl esi, 16
nooverride:
		push esi
		push ecx
		mov edi, esi
		mov esi, offset SAMPLEBUFFER
		rep movsb
		pop ecx
		pop esi

		mov ebx, esi
		shr ebx, 16

;
; Setup DMA-controller :
;
; 1st  MASK DMA CHANNEL
;
           mov     al,DMAchannel
           add     al,4
           out     WRITEMASK,al
;
; 2nd  CLEAR FLIPFLOP
;
           out     CLEARFLIPFLOP,al
;
; 3rd  WRITE TRANSFER MODE
;
           mov     al,WANTEDMODE
           add     al,DMAchannel
           out     WRITEMODE,al
;
; 4th  WRITE PAGE NUMBER
;
           mov     al,bl
           out     PAGE_CHN,al
;
; 5th  WRITE BASEADDRESS
;
           mov     ax,si
           out     BASE_CHN,al
           mov     al,ah
           out     BASE_CHN,al
;
; 6th  WRITE SAMPLELENGTH-1
;
           mov     al,cl
           out     COUNT_CHN,al
           mov     al,ch
           out     COUNT_CHN,al
;
; 7th  DEMASK CHANNEL
;
           mov     al,DMAchannel
           out     WRITEMASK,al

;
; Setup SoundBlaster :
;
; 1st  SET TIMECONSTANTE
;
           mov     dx,BASEADDR+00Ch            ;DX = DSP Write Data or Command
           WAITWRITE
           mov     al,040h                     ;AL = Set timeconstant
           out     dx,al
           WAITWRITE
           mov     al,TIMECONST
           out     dx,al
;
; 2nd  use 8bit mono nonautoinit (DSPcommand 014h XXXXh)
;
           WAITWRITE
           mov     al,014h                     ;AL = DMA DAC 8bit
           out     dx,al
           mov     cx,SAMPLEBUFFERLENGTH-1
           WAITWRITE
           mov     al,cl                       ;AL = LOWER PART SAMPLELENGTH
           out     dx,al
           WAITWRITE
           mov     al,ch                       ;AL = HIGHER PART SAMPLELENGTH
           out     dx,al

; TRANSFER STARTS ....... NOW ..... :)

waitloop:
           mov     ah,01                       ;AH = Check for character function
           int     016h                        ;   Interrupt: Keyboard
           jnz     DMAstop
           cmp     [ready],0                   ; check if sound sent
           jz      waitloop

exit:      ; RESTORE PIC MASK
           in      al,021h
           or      al,PICORMASK                ;<-- SET REGISTER MASK BITS TO DISABLE
           out     021h,al

           ; RESTORE IRQ :
		   mov edx, [OLDInterruptOFS]
		   mov  cx, [OLDInterruptSEG]
		   mov bl, SBIRQ
		   mov ax, 205h
		   int 31h

           ; CLEAR KEYBUFFER
           mov     ah,01                       ;AH = Check for character function
           int     16h                         ;   Interrupt: Keyboard
           jz      @F
           xor     ah,ah                       ;Read character, flush keypress
           int     016h                        ;   Interrupt: Keyboard
@@:
           ret

DMAstop:
           RESET_DSP
           JMP     exit

main endp

dosmemerr:
           mov edx, offset dosmemerror
		   mov ah,9
		   int 21h
		   jmp return2dos

; display information if Soundblaster is not on this baseaddress

RESET_ERROR:
           mov     edx,offset sberror
           mov     ah,9
           int     21h                         ; text output
           jmp     return2dos


;
; Our own IRQ for detecting end of playing
; It's generated by the SoundBlaster hardware
;
OWN_IRQ:   push    eax
           push    edx
           push    ds
		   mov ds, cs:[dataseg]
           mov     dx,BASEADDR+00Eh            ;DX = DSP DATA AVAILABLE (IRQ ACKNOWLEDGE)
           in      al,dx
           mov     [READY],1
           mov     al,020h
           out     020h,al                     ;ACKNOWLEDGE HARDWARE INTERRUPT
           pop     ds
           pop     edx
           pop     eax
           IRETD

_start32:
	call main
return2dos:
	mov ah, 4Ch
	int 21h

	END _start32
